#Copyright (C) 2005 by Dr. Dieter Maurer <dieter@handshake.de> -- see LICENSE.txt
#       $Id: pypackage.py,v 1.2 2005/05/27 12:09:14 dieter Exp $
'''Access to 'pypackage:' URLs and normal file references.'''

import os, os.path, __builtin__, glob as pyglob

from os.path import sep
from os import error
from sys import exc_info
from stat import S_IFREG, S_IFDIR
from types import StringTypes
from time import mktime

from util import getBuiltin


__all__ = ['join', 'split', 'getPackageUrl',
           'open', 'stat', 'listdir', 'glob',
           'getLocation',
           ]


# forming and parsing package URLs

def join(package, path):
    '''form package URL from package name *package* and *path*.'''
    # normalize
    if sep != '/': path = path.replace(sep, '/')
    return 'pypackage://%s' % package + (path and '/'+path or '')


def split(url):
    '''split *url* into package name and path.'''
    if not url.startswith('pypackage:'):
        raise ValueError('not a pypackage url: %s' % url)
    url = url[10:]
    if sep != '/': url = url.replace(sep, '/')
    if url[:2] == '//': url = url[2:]
    parts  = url.split('/', 1)
    package, path = len(parts) == 2 and parts or (parts[0], '')
    return _normalize(package, path)


def getPackageUrl(module):
    '''the package URL for *module*.

    *module* may either be a module or its dictionary.

    If *module* is not a package, its enclosing package is used.
    '''
    if not isinstance(module, dict): module = module.__dict__
    name = module['__name__']
    if module.get('__path__') is None:
        # not a package
        pname = '.'.join(name.split('.')[:-1])
        if not pname:
            raise ValueError(
                'a top level module does not have a package URL: %s'
                % name
                )
        name = pname
    return join(name, '')


# the primary access functions

def open(file, *args, **kw):
    if not file.startswith('pypackage:'):
        return _builtin__open(file, *args, **kw)
    if args: mode = args[0]
    else: mode = kw.get('mode', 'r')
    if mode not in ('r', 'rb', 'rt'):
        raise TypeError("mode %s not supported for 'pypackage' access" % mode)
    try: return _lookupResource(file).open(mode)
    except OSError, e:
        # convert to an IOError (as this is what 'open' usually reports)
        raise IOError(e.errno, e.strerror, e.filename), None, exc_info()[2]


def stat(file):
    if not file.startswith('pypackage:'): return _builtin__stat(file)
    return _lookupResource(file).stat()


def getLocation(file):
    '''return something similar to a filesystem location (if possible).

    This function is primarily used to obtain a value passed as
    'filename' for compilations. Using a true filesystem location
    gives debuggers a chance to make such objects debuggable.
    '''
    if not file.startswith('pypackage:'): return file
    return _lookupResource(file).getLocation()


def listdir(file):
    if not file.startswith('pypackage:'): return _builtin__listdir(file)
    return _collectResources(file, 'listdir')


def glob(pattern):
    if not pattern.startswith('pypackage:'):
        return _builtin__glob(pattern)
    package, pattern = split(pattern)
    purl = join(package, '')
    prefix = purl + '/'
    return [prefix + path for path in _collectResources(purl, 'glob', pattern)]


# more 'os' functions handling pypackage URLs
# Note: we might want more functions in the future (e.g. 'os.walk')

def execfile(file, *args):
    if not args:
        # Ups, the nasty form of 'execfile' which should use and modify
        # the local namespace of the caller
        from sys import _getframe
        frame = _getframe(1)
        args = frame.f_globals, frame.f_locals
    if not file.startswith('pypackage:'):
        return _builtin__execfile(file, *args)
    if len(args) == 1: globals = locals = args[0]
    else: globals, locals = args
    code = compile(open(file, 'r').read(), getLocation(file), 'exec')
    exec code in globals, locals


# Customization

_frozenLookupPath = None

def getFrozenLookupPath(package):
    '''return lookup path for frozen *package*.

    This implementation ignores *package* and unconditionally
    returns the path given by the envvar 'PACKAGERESOURCES_FROZEN_LOOKUP_PATH'.
    '''
    global _frozenLookupPath
    lookup_path = _frozenLookupPath
    if lookup_path is None: 
        from os import environ
        from os.path import pathsep
        pathstring = environ.get('PACKAGERESOURCES_FROZEN_LOOKUP_PATH')
        lookup_path = _frozenLookupPath = \
                      pathstring and pathstring.split(pathsep) or []
    return lookup_path


# Auxiliaries

def _lookupResource(pypurl):
    resources = _lookupResources(pypurl)
    try: return resources.next()
    except StopIteration:
        raise OSError(2, 'No such resource', pypurl)

def _lookupResources(pypurl, u_=({},{},['__doc__'])):
    '''generator enumerating all resources corresponding to *pypurl*.'''
    package, path = split(pypurl)
    lookup_path = __import__(package, *u_).__path__
    if isinstance(lookup_path, StringTypes):
        # frozen package
        lookup_path = getFrozenLookupPath(package)
    from sys import path_hooks, path_importer_cache
    for path_item in lookup_path:
        importer = path_importer_cache.get(path_item, 0)
        if importer == 0:
            # see which importer type is responsible
            # this code emulates "import.c"
            for hook in path_hooks:
                try: importer = hook(path_item)
                except ImportError: pass
                else: break
            if importer == 0:
                # no hook can process this item -- use default
                importer = None
            path_importer_cache[path_item] = importer
        factory = _makeResourceFactory(importer, path_item)
        # we silently ignore importers we do not understand
        resource = factory and factory(path)
        if resource is not None: yield resource


def _collectResources(file, method, *args):
    seen = {}; result = []
    for resource in _lookupResources(file):
        for name in getattr(resource, method)(*args):
            if name in seen: continue
            seen[name] = None
            result.append(name)
    return result


def _makeResourceFactory(importer, path_item):
    '''turn *importer* into a resource factory, if we understand it.

    Currently, we understand only directory and zip importers.

    We return 'None' for importers we do not understand.
    '''
    if importer is None: return _FileSystemResourceFactory(path_item)
    if hasattr(importer, '_files') \
       and hasattr(importer, 'archive') \
       and hasattr(importer, 'prefix') \
       and hasattr(importer, 'get_data'):
        # looks sufficiently like a zip importer that we try it.
        return _ZipResourceFactory(importer)


class _FileSystemResourceFactory:
    def __init__(self, base):
        self.base = base

    def __call__(self, path):
        if sep != '/': path = path.replace('/', sep)
        base = self.base
        full_path = path and _builtin__join(base, path) or base
        try: stat_result = _builtin__stat(full_path)
        except OSError: return
        return _FileSystemResource(full_path, stat_result)


class _ZipResourceFactory:
    def __init__(self, importer):
        self.archive = self._getArchive(importer)
        prefix = importer.prefix
        if sep != '/': prefix = prefix.replace(sep, '/')
        self.prefix = prefix

    def __call__(self, path):
        archive = self.archive
        full_path = self.prefix + path
        if full_path.endswith('/'): full_path = full_path[:-1]
        if full_path in archive: return _ZipResource(archive, full_path)

    _archives = {}
    def _getArchive(class_, importer):
        name = importer.archive
        archive = class_._archives.get(name)
        if archive is None:
            archive = class_._archives[name] = _ZipArchive(name)
        return archive
    _getArchive = classmethod(_getArchive)


class _FileSystemResource:
    def __init__(self, full_path, stat):
        self.full_path = full_path
        self._stat = stat

    def stat(self): return self._stat

    def open(self, mode): return _builtin__open(self.full_path, mode)

    def listdir(self): return _builtin__listdir(self.full_path)

    def glob(self, pattern):
        join = _builtin__join; normalize = sep != '/'
        full_path = self.full_path; pl = len(full_path) + 1
        for path in _builtin__glob(join(full_path, pattern)):
            path = path[pl:]
            if normalize: path = path.replace(sep, '/')
            yield path

    def getLocation(self): return self.full_path


class _ZipResource:
    def __init__(self, archive, full_path):
        self.archive = archive
        self.full_path = full_path

    def stat(self):
        info = self.archive[self.full_path]
        path, cmode, dsize, fsize, foffset, mtime, mdate, crc = info
        st_mode = 0444 + (foffset and S_IFREG or S_IFDIR)
        st_ino = st_dev = st_uid = st_gid = 0
        st_nlink = 1
        st_size = fsize
        mtime = _convDosTime(mdate, mtime)
        st_atime = st_mtime = st_ctime = mtime
        return _Stat(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime)

    def listdir(self):
        return self.archive.listdir(self.full_path)

    def glob(self, pattern):
        return self.archive.glob(self.full_path, pattern)

    def open(self, mode):
        data = self.archive.get_data(self.full_path)
        from sys import platform
        from cStringIO import StringIO
        if mode != 'rb':
            if platform == 'win32': data = data.replace('\r\n', '\n')
            # something is missing here for Mac
        return StringIO(data)

    def getLocation(self):
        return self.archive[self.full_path][0]


class _ZipArchive:
    def __init__(self, name):
        from zipimport import zipimporter
        i = zipimporter(name)
        self.name = name
        self.get_data = i.get_data
        info = self._info = {}
        normalize = sep != '/'
        for file, finfo in i._files.iteritems():
            if normalize: file = file.replace(sep, '/')
            if file.endswith('/'):
                # a folder
                file = file[:-1]
                if finfo[4]:
                    # "zipimport" promisses this would not happen
                    finfo = finfo[:4] + (0,) + finfo[5:]
            info[file] = finfo
        # Our use requires that all directories are present.
        # Some zip archivers (e.g. the Python one) do not save
        # directories. Thus create fakes for missing directories.
        # Also create a hierarchical map.
        fake_dir_info = (0, 0, 0, 0, 0, 0, 0); normalize = sep != '/'
        from sys import maxint
        hmap = self._hmap = {}
        for file in info.keys():
            epos = maxint
            while epos > 0:
                spos = file.rfind('/', 0, epos)
                dir, fname = spos != -1 and (file[:spos], file[spos+1:epos]) or ('', file[:epos])
                content = hmap.get(dir)
                if content is None: content = hmap[dir] = []
                content.append(fname)
                if dir in info: break
                fdir = dir
                if normalize: fdir = fdir.replace('/', sep)
                info[dir] = (_builtin__join(name, fdir),) + fake_dir_info
                epos = spos
                
    def listdir(self, path): return self._hmap[path]

    def glob(self, path, pattern):
        if not pattern: raise StopIteration
        from fnmatch import filter
        pattern = pattern.split('/')
        hmap = self._hmap
        def rglob(path, pattern):
            comp_pat = pattern.pop()
            base = not pattern and (path,) or rglob(path, pattern)
            for ppath in base:
                names = hmap.get(ppath)
                if names:
                    for name in filter(names, comp_pat):
                        yield ppath and ppath + '/' + name or name
        pl = len(path)
        for ppath in rglob(path, pattern): yield ppath[pl:]

    def __getitem__(self, key):
        return self._info[key]

    def __contains__(self, key):
        return key in self._info

    def __iter__(self):
        return iter(self._info)


class _Stat:
    '''wrapper for 'stat' result to make 'os.path' functions happy.'''
    def __init__(self, *values):
        self._tuple = values

    def __getitem__(self, i): return self._tuple[i]

    def __getattr__(self, k):
        names = self._names
        if k not in  names: raise AttributeError(k)
        return self._tuple[names[k]]

    _names = dict([(n,i) for (i,n) in enumerate('st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime'.split(', '))])
    

def _convDosTime(ddate, dtime):
    '''convert DOS *ddate*, *dtime* (two shorts) into sec since epoch.
    '''
    year = (ddate >> 9) + 1980
    month = (ddate >> 5) & 0xf
    day = ddate & 0x1f
    hour = dtime >> 11
    min = dtime >> 5 & 0x3f
    sec = (dtime & 0x1f) << 1
    return int(mktime((year, month, day, hour, min, sec, 0, 0, 0)))


def _normalize(package, path):
    spath = path.split('/')
    di = ci = 0
    while ci < len(spath):
        oci = ci; ci += 1
        seg = spath[oci]
        if not seg or seg == '.': continue # forget empty segments
        if seg == '..':
            # parent indicator
            if di: di -= 1
            else:
                # strip last package component
                pos = package.rfind('.')
                if pos == -1:
                    raise ValueError('.. leaves package context')
                package = package[:pos]
            continue
        if oci != di: spath[di] = seg
        di += 1
    if di == ci: return package, path
    return package, '/'.join(spath[:di])


# import originals (not patched)

_builtin__open = getBuiltin(__builtin__, 'open')
_builtin__execfile = getBuiltin(__builtin__, 'execfile')
_builtin__stat = getBuiltin(os, 'stat')
_builtin__listdir = getBuiltin(os, 'listdir')
_builtin__join = getBuiltin(os.path, 'join')
_builtin__glob = getBuiltin(pyglob, 'glob')

