#Copyright (C) 2005 by Dr. Dieter Maurer <dieter@handshake.de> -- see LICENSE.txt
#       $Id: test_packageresources.py,v 1.1 2005/05/27 10:21:52 dieter Exp $

from unittest import TestCase, makeSuite, TestSuite, TextTestRunner
from os.path import dirname, join, abspath
from os import getcwd, path, environ
from sys import modules
import os, glob, __builtin__

from packageresources import getPackageUrl, getLocation, pypackage
import packageresources.path

_base = abspath(dirname(__file__))

# patch 'testpackage.__path__' to include 'test.zip'
zip = join(_base, 'test.zip')
import testpackage; testpackage.__path__.append(zip)
testurl = getPackageUrl(testpackage)


class _ResourceTests(TestCase):
    # to be defined by derives classes
    #PREFIX
    #BLT
    #PATH
    #OS
    #GLOB

    def test_open(self):
        prefix = self.PREFIX; open = self.BLT.open
        data = open(testurl + '/' + prefix + 'm.py').read()
        self.assertEqual(data, prefix + ' = 1\n')
        self.assertRaises(IOError, open, testurl + '/x') # does not exist
        self.assertRaises(IOError, open, testurl) # is a directory

    def test_classification(self):
        path = self.PATH; prefix = self.PREFIX
        self.assert_(path.isdir(join(testurl, prefix + 'd1')))
        self.assert_(path.isdir(join(testurl, prefix + 'd2')))
        self.assert_(not path.isdir(join(testurl, prefix + 'd1/x.py')))
        self.assert_(path.isfile(join(testurl, prefix + 'd1/x.py')))
        self.assert_(not path.isfile(join(testurl, prefix + 'd1/y')))

    def test_listdir(self):
        listdir = self.OS.listdir; prefix = self.PREFIX
        l = listdir(testurl)
        self._compatible(l, ['__init__.py', 'd', 'fd1', 'fd2', 'fm.py', 'zd1', 'zd2', 'zm.py'])
        self._compatible(listdir(join(testurl, prefix + 'd2')), ['y'])
        l = listdir(testurl + '.' + prefix + 'd1')
        self._compatible(l, ['__init__.py', 'x.py'])
        l = listdir(join(testurl, 'd'))
        self._compatible(l, ['fz', 'z', 'zz'])

    def test_glob(self):
        glob = self.GLOB.glob; prefix = self.PREFIX
        l = glob(testurl + '/*/y')
        self._compatible(l, [testurl + '/fd2/y', testurl + '/zd2/y'])
        pref = testurl + '/' + prefix + 'd1/'
        self._compatible(glob(pref + '_*'), [pref + '__init__.py'])
        pref = testurl + '/d/'
        l = glob(pref + '*z')
        self._compatible(l, [pref + 'fz', pref + 'z', pref + 'zz'])

    def test_getLocation(self):
        prefix = self.PREFIX
        bn = prefix == 'f' and 'testpackage' or 'test.zip'
        self.assertEqual(getLocation(testurl + '/' + prefix + 'm.py'),
                         join(_base, bn, prefix + 'm.py')
                         )

    def test_execfile(self):
        prefix = self.PREFIX; execfile = self.BLT.execfile
        file = testurl + '/' + prefix + 'm.py'
        g = {}
        execfile(file, g)
        self.assertEqual(g[prefix], 1)
        g = {}; l = {}
        execfile(file, g, l)
        self.assertEqual(l[prefix], 1)
        g = {'execfile':execfile, 'file':file}; l = {}
        exec "execfile(file)" in g, l
        self.assertEqual(l[prefix], 1)

    def _compatible(self, be, should):
        '''check that *be* is compatible with *shouls*.'''
        # remove any CVS reference, '.pyc', '.pyo' and '.log' files
        be = [item for item in be if
              'CVS' not in item
              and not item.endswith('.pyc')
              and not item.endswith('.pyo')
              and not item.endswith('.log')
              ]
        # sort, as order is unessential (and 'should' is ordered)
        be.sort()
        self.assertEqual(be, should)


class _PathTests(TestCase):
    # to be defined by derives classes
    #PATH

    def test_isabs(self):
        isabs = self.PATH.isabs
        self.assert_(isabs(testurl))
        self.assert_(not isabs('x'))

    def test_abspath(self):
        abspath = self.PATH.abspath
        self.assertEqual(abspath(testurl), testurl)
        self.assertEqual(abspath('x'), join(getcwd(), 'x'))

    def test_join(self):
        join = self.PATH.join
        self.assertEqual(join('a', testurl), testurl)
        self.assertEqual(join(testurl, 'a'), testurl + '/a')
        cwd = getcwd()
        self.assertEqual(join(testurl, cwd), cwd)
        self.assertEqual(join(testurl, testurl, 'a'), testurl + '/a')
        self.assertEqual(join(cwd, testurl), testurl)

    def test_normpath(self):
        normpath = self.PATH.normpath
        self.assertEqual(normpath(testurl + '/'), testurl)
        self.assertEqual(normpath(join(testurl, 'a')), testurl + '/a')
        self.assertEqual(normpath(join(testurl, 'a', '..')), testurl)
        self.assertEqual(normpath(join(testurl+'.x', '..')), testurl)

class UrlTests(TestCase):
    def test_getPackageUrl(self):
        from testpackage import fm, zm
        self.assertEqual(getPackageUrl(fm), testurl)
        self.assertEqual(getPackageUrl(zm.__dict__), testurl)

class FrozenTests(TestCase):
    fmid = 'frozen_test_module'
    fmurl = 'pypackage://' + fmid

    def setUp(self):
        from types import ModuleType
        id = self.fmid
        frozen = ModuleType(id)
        frozen.__path__ = '<frozen>'
        modules[id] = frozen
        environ['PACKAGERESOURCES_FROZEN_LOOKUP_PATH'] = zip

    def tearDown(self):
        del modules[self.fmid]

    def test_access(self):
        packageresources.path.isdir(self.fmurl + '/d')


## Mixin classes
class _FSTests: PREFIX = 'f'
class _ZIPTests: PREFIX = 'z'

class _Patched:
    OS = os
    GLOB = glob
    BLT = __builtin__
    PATH = path

    def setUp(self):
        from packageresources import patchPython
        patchPython()

    def tearDown(self):
        from packageresources.MonkeyPatch import unpatchPython
        unpatchPython()

class _Unpatched:
    OS = GLOB = BLT = pypackage
    from packageresources import path as PATH

## Combined test classes

class PatchedFSTests(_FSTests, _Patched, _ResourceTests,): pass
class PatchedZIPTests(_ZIPTests, _Patched, _ResourceTests,): pass
class PatchedPathTests(_Patched, _PathTests,): pass
class UnpatchedFSTests(_FSTests, _Unpatched, _ResourceTests,): pass
class UnpatchedZIPTests(_ZIPTests, _Unpatched, _ResourceTests,): pass
class UnpatchedPathTests(_Unpatched, _PathTests,): pass



def test_suite():
    return TestSuite([makeSuite(c) for c in (
        UrlTests,
        FrozenTests,
        PatchedFSTests,
        PatchedZIPTests,
        PatchedPathTests,
        UnpatchedFSTests,
        UnpatchedZIPTests,
        UnpatchedPathTests,
        )]
                     )

def run_tests():
    runner = TextTestRunner()
    runner.run(test_suite())


if __name__ == '__main__':
    run_tests()
